from paretoset import paretoset
import numpy as np


import pytest
import pandas as pd
import itertools


allowed_min_values = [min, "min", "MIN", "minimum", "Min"]
allowed_max_values = [max, "max", "MAX", "maximum", "Max"]
bools = [True, False]


class TestReadmeExamples:
    def test_example_hotels(self):
        """Example in the readme."""
        from paretoset import paretoset
        import pandas as pd

        hotels = pd.DataFrame(
            {"price": [50, 53, 62, 87, 83, 39, 60, 44], "distance_to_beach": [13, 21, 19, 13, 5, 22, 22, 25]}
        )
        mask = paretoset(hotels, sense=["min", "min"])
        effi_hotels = hotels[mask]
        effi_hotels = effi_hotels

    def test_example_salespeople(self):
        """Example in the readme."""
        from paretoset import paretoset
        import pandas as pd

        salespeople = pd.DataFrame(
            {
                "salary": [94, 107, 67, 87, 153, 62, 43, 115, 78, 77, 119, 127],
                "sales": [123, 72, 80, 40, 64, 104, 106, 135, 61, 81, 162, 60],
                "department": ["c", "c", "c", "b", "b", "a", "a", "c", "b", "a", "b", "a"],
            }
        )
        mask = paretoset(salespeople, sense=["min", "max", "diff"])
        top_performers = salespeople[mask]
        top_performers = top_performers

    def test_example_optimization(self):
        """Example in the readme."""
        from paretoset import paretoset
        import numpy as np
        from collections import namedtuple

        np.random.seed(42)

        # Create Solution objects holding the problem solution and objective values
        Solution = namedtuple("Solution", ["solution", "objective_values"])
        solutions = [Solution(solution=object, objective_values=np.random.randn(2)) for _ in range(999)]

        # Create an array of shape (solutions, objectives) and compute the non-dominated set
        objective_values_array = np.vstack([s.objective_values for s in solutions])
        mask = paretoset(objective_values_array, sense=[min, min])

        # Filter the list of solutions, keeping only the non-dominated solutions
        efficient_solutions = [solution for (solution, m) in zip(solutions, mask) if m]
        efficient_solutions = efficient_solutions

    def test_example_several_different(self):
        """A small example worked by hand."""

        df = pd.DataFrame(
            {
                "col1": ["a", "a", "a", "a", "a", "b", "b", "b", "b"],
                "col2": ["A", "A", "B", "B", "B", "A", "B", "B", "B"],
                "col3": [-1, 0, 0, 1, 1, -2, 1, 2, 1],
                "col4": [1, -1, -1, -1, 0, 2, 0, 0, 0],
            }
        )
        sense = ["diff", "diff", "min", "max"]
        mask = paretoset(df, sense=sense, distinct=True)
        expected = np.array([1, 0, 1, 0, 1, 1, 1, 0, 0], dtype=np.bool_)
        assert np.all(mask == expected)

        mask = paretoset(df, sense=sense, distinct=False)
        expected = np.array([1, 0, 1, 0, 1, 1, 1, 0, 1], dtype=np.bool_)
        assert np.all(mask == expected)

    @pytest.mark.parametrize("operator, distinct", list(itertools.product(["min", "max"], bools)))
    def test_on_non_numeric_data(self, operator, distinct):
        """Test that non-numeric data fails."""

        df = pd.DataFrame({"col1": [0, 1, 0, 1], "col2": ["A", "A", "B", "B"]})

        # This will not work because 'operator' is applied over non-numeric data
        with pytest.raises(TypeError):
            paretoset(df, sense=["diff", operator], distinct=distinct)

        # Same as above
        with pytest.raises(TypeError):
            paretoset(df, sense=[operator, operator], distinct=distinct)

        # Fails on NumPy arrays too
        with pytest.raises(TypeError):
            data = np.array([[1, 1], [0, 1]], dtype=object)
            paretoset(data, sense=[operator, operator], distinct=distinct)

        # This will work, because the operator works on numeric data
        mask = paretoset(df, sense=[operator, "diff"], distinct=distinct)
        assert np.any(mask)

    @pytest.mark.parametrize("dtype", [int, float])
    def test_diff_on_numeric_data(self, dtype):
        """Test that 'diff' works on numeric data."""
        # Generate random data
        np.random.seed(42)
        costs = np.random.randint(low=-2, high=2, size=(99, 2))
        costs = np.asarray(costs, dtype=dtype)

        assert np.any(paretoset(costs, sense=["diff", "min"]))


class TestParetoSetAPI:
    @pytest.mark.parametrize(
        "dtype, distinct, use_numba", list(itertools.product([np.asarray, pd.DataFrame], bools, bools))
    )
    def test_at_least_one_chosen(self, dtype, distinct, use_numba):
        """Test that the input is not mutated."""

        # Generate random data
        np.random.seed(42)
        costs = np.random.randint(low=-2, high=2, size=(99, 3))

        # Get the mask
        mask = paretoset(dtype(costs), sense=[min, max, min], distinct=distinct, use_numba=use_numba)

        # Verify that at least one element is chosen
        assert np.any(mask)

    @pytest.mark.parametrize(
        "dtype, distinct, use_numba", list(itertools.product([np.asarray, pd.DataFrame], bools, bools))
    )
    def test_no_side_effects(self, dtype, distinct, use_numba):
        """Test that the input is not mutated."""

        # Generate random data
        np.random.seed(42)
        costs = np.random.randint(low=-2, high=2, size=(99, 3))
        costs = dtype(costs)

        # Copy the data before passing into function
        costs_before = costs.copy()
        paretoset(costs, sense=[min, max, min], distinct=distinct, use_numba=use_numba)

        # The input argument is not changed/mutated in any way
        assert costs_before.shape == costs.shape
        assert np.all(costs_before == costs)

    @pytest.mark.parametrize("sense_min, sense_max", list(itertools.product(allowed_min_values, allowed_max_values)))
    def test_sense_argument(self, sense_min, sense_max):
        """Test the synonyms for the `sense` argument."""

        np.random.seed(42)
        costs = np.random.randint(low=-2, high=2, size=(99, 3))

        mask1 = paretoset(costs, sense=["min", "max", "diff"])
        mask2 = paretoset(costs, sense=[sense_min, sense_max, "diff"])

        # Masks are equal
        assert np.all(mask1 == mask2)

    def test_max_and_diff_working_together(self):
        """Regression test for:
        https://github.com/tommyod/paretoset/issues/18
        """
        mask = paretoset([[2, 0], [1, 0], [0, 1]], sense=["max", "diff"])
        assert np.all(mask == np.array([True, False, True]))


class TestPandasInputs:
    def test_pandas_no_args(self):
        costs = np.random.RandomState(123).randn(5, 2)
        df = pd.DataFrame(costs, columns=["a", "b"])
        df_before = df.copy()
        mask = paretoset(df)

        assert isinstance(mask, np.ndarray)
        assert mask.dtype == bool
        assert (df == df_before).all().all()

    def test_pandas_with_args(self):
        costs = np.random.RandomState(123).randn(5, 2)
        df = pd.DataFrame(costs, columns=["a", "b"])
        df_before = df.copy()
        mask = paretoset(df, sense=[min, max])

        assert isinstance(mask, np.ndarray)
        assert mask.dtype == bool
        assert (df == df_before).all().all()

    def test_pandas_with_diff_arg(self):
        costs = np.random.RandomState(123).randn(24, 2)
        df = pd.DataFrame(costs, columns=["a", "b"])
        df["department"] = ["a", "b", "c"] * 8
        df_before = df.copy()
        mask = paretoset(df, sense=["min", "max", "diff"])

        assert isinstance(mask, np.ndarray)
        assert mask.dtype == bool
        assert (df == df_before).all().all()


if __name__ == "__main__":
    pytest.main(args=[__file__, "--doctest-modules", "--maxfail=5", "-v", "--cache-clear", "--color", "yes", ""])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Tests for algorithms related to association rules.
"""

from paretoset.algorithms_numpy import paretoset_naive, paretoset_efficient, pareto_rank_naive, crowding_distance
from paretoset.algorithms_numba import paretoset_jit, BNL

import pytest
import numpy as np
import itertools


seeds = list(range(99))
dtypes = [float, int]
bools = [True, False]
paretoset_algorithms = [paretoset_naive, paretoset_efficient, paretoset_jit, BNL]
paretorank_algorithms = [pareto_rank_naive]


def generate_problem_simplex(n, d):
    """Generate D dimensional data on the D-1 dimensional simplex."""
    # https://cs.stackexchange.com/questions/3227/uniform-sampling-from-a-simplex
    data = np.random.randn(n, d - 1)
    data = np.hstack((np.zeros(n).reshape(-1, 1), data, np.ones(n).reshape(-1, 1)))
    diffs = data[:, 1:] - data[:, :-1]
    assert diffs.shape == (n, d)
    return diffs


class TestParetoSetImplementations:
    @pytest.mark.parametrize("seed, dtype, distinct", itertools.product(seeds, dtypes, bools))
    def test_on_random_instances(self, seed, dtype, distinct):
        """Test that the algorithms all return the same answer."""

        # Generate a random instance
        np.random.seed(seed)
        n_costs = np.random.randint(1, 99)
        n_objectives = np.random.randint(1, 4)
        costs = np.random.randn(n_costs, n_objectives)

        # Convert to dtype, this creates some duplicates when `dtype` is integer
        costs = np.array(costs, dtype=dtype)

        # Compute the answers
        masks = [algo(costs, distinct=distinct) for algo in paretoset_algorithms]

        # At least one element must be in the mask
        assert all(np.sum(m) > 0 for m in masks)

        # Check pairwise that the answers are identical
        for m1, m2 in zip(masks[:-1], masks[1:]):
            assert np.all(m1 == m2)

    @pytest.mark.parametrize("algorithm", paretoset_algorithms)
    def test_case_distinct_1(self, algorithm):
        """Test the `distinct` parameter on a simple example"""
        costs = np.array([[1, 1], [0, 1], [0, 1]])

        ranks_distinct = algorithm(costs, distinct=True)
        assert np.all(ranks_distinct == np.array([False, True, False]))

        ranks_non_distinct = algorithm(costs, distinct=False)
        assert np.all(ranks_non_distinct == np.array([False, True, True]))

    @pytest.mark.parametrize("algorithm", paretoset_algorithms)
    def test_case_distinct_2(self, algorithm):
        """Test the `distinct` parameter on a simple example"""
        costs = np.array([[1, 0], [1, 0], [0, 1], [0, 1]])

        ranks_distinct = algorithm(costs, distinct=True)
        assert np.all(ranks_distinct == np.array([True, False, True, False]))

        ranks_non_distinct = algorithm(costs, distinct=False)
        assert np.all(ranks_non_distinct)

    @pytest.mark.parametrize("algorithm", paretoset_algorithms)
    def test_case_distinct_3(self, algorithm):
        """Test the `distinct` parameter on a simple example"""
        costs = np.array([[1, 0], [1, 0], [0, 1], [0, 1], [0.5, 0.5], [0.5, 0.5]])

        ranks_distinct = algorithm(costs, distinct=True)
        assert np.all(ranks_distinct == np.array([True, False, True, False, True, False]))

        ranks_non_distinct = algorithm(costs, distinct=False)
        assert np.all(ranks_non_distinct)

    @pytest.mark.parametrize("algorithm", paretoset_algorithms)
    def test_case_distinct_4(self, algorithm):
        """Test the `distinct` parameter on a simple example"""
        costs = np.array([[0], [0]])

        ranks_distinct = algorithm(costs, distinct=True)
        assert np.all(ranks_distinct == np.array([True, False]))

        ranks_non_distinct = algorithm(costs, distinct=False)
        assert np.all(ranks_non_distinct)

    @pytest.mark.parametrize("seed, algorithm", itertools.product(seeds, paretoset_algorithms))
    def test_invariance_under_permutations(self, seed, algorithm):
        """Test that the algorithm in invariant under random permutations of data."""

        # Create some random data
        np.random.seed(seed)
        n_costs = np.random.randint(1, 9)
        n_objectives = np.random.randint(1, 4)
        costs = np.random.randint(low=-1, high=1, size=(n_costs, n_objectives))

        # Get masks
        mask_distinct = algorithm(costs, distinct=True)
        ranks_non_distinct = algorithm(costs, distinct=False)

        # Permute the data
        permutation = np.random.permutation(np.arange(n_costs))
        assert np.sum(mask_distinct) > 0
        assert np.sum(ranks_non_distinct) > 0

        # When `distinct` is set to `False`, permutation invariance should hold
        assert np.all(ranks_non_distinct[permutation] == algorithm(costs[permutation], distinct=False))

        # Equally many should me marked in the mask, regardless of `distinct` or not
        assert np.sum(mask_distinct[permutation]) == np.sum(algorithm(costs[permutation], distinct=True))
        assert np.sum(ranks_non_distinct[permutation]) == np.sum(algorithm(costs[permutation], distinct=False))

    @pytest.mark.parametrize("seed, algorithm", itertools.product(seeds, paretoset_algorithms))
    def test_equal_values(self, seed, algorithm):
        """For each group of identical data in the Pareto set: if `distinct`,
        the first index should be True and everything else False.
        If not `distinct`, the group should all be True.

        """

        # Generate random data
        np.random.seed(seed)
        n_costs = np.random.randint(1, 99)
        n_objectives = np.random.randint(1, 4)
        costs = np.random.randint(low=-1, high=1, size=(n_costs, n_objectives))

        indices = np.arange(n_costs)

        # Go through every cost in the Pareto set
        mask_distinct = algorithm(costs, distinct=True)
        for cost in costs[mask_distinct]:
            mask_inds = indices[np.all(cost == costs, axis=1)]

            # Only one marked
            assert np.sum(mask_distinct[mask_inds]) == 1

            # The marked element is the first one
            assert mask_distinct[mask_inds][0]

        # Go through every cost in the Pareto set
        mask_unique = algorithm(costs, distinct=False)
        for cost in costs[mask_unique]:
            mask_inds = indices[np.all(cost == costs, axis=1)]

            # All are marked
            assert np.all(mask_unique[mask_inds])

    @pytest.mark.parametrize("algorithm, distinct", itertools.product(paretoset_algorithms, bools))
    def test_no_side_effects(self, algorithm, distinct):
        """Test that the input is not mutated."""

        # Generate random data
        np.random.seed(42)
        costs = np.random.randint(low=-2, high=2, size=(99, 3))

        # Copy the data before passing into function
        costs_before = costs.copy()
        algorithm(costs, distinct=distinct)

        # The input argument is not changed/mutated in any way
        assert costs_before.shape == costs.shape
        assert np.all(costs_before == costs)


class TestParetoRankImplementations:
    @pytest.mark.parametrize("algorithm, seed", itertools.product(paretorank_algorithms, seeds))
    def test_paretorank_on_random_instances(self, algorithm, seed):
        """"""
        # Generate random data
        np.random.seed(42)
        costs = np.random.randint(low=-2, high=2, size=(99, 3))

        ranks = algorithm(costs)

        # Minimum rank is 1, no rank greater than the number of rows
        assert np.min(ranks) == 1
        assert np.max(ranks) <= len(ranks)

    @pytest.mark.parametrize("algorithm", paretorank_algorithms)
    def test_paretorank_on_degenerate_case(self, algorithm):
        """"""
        # Generate random data
        costs = np.ones((9, 1))

        # Ranks are 1, 1, 1, 1, ...
        ranks = algorithm(costs, distinct=False)
        assert np.all(ranks == 1)

        # Ranks are 1, 2, 3, 4, ...
        ranks = algorithm(costs, distinct=True)
        assert np.all(ranks == (np.arange(len(ranks)) + 1))

    @pytest.mark.parametrize("seed, algorithm", itertools.product(seeds, paretorank_algorithms))
    def test_invariance_under_permutations(self, seed, algorithm):
        """Test that the algorithm in invariant under random permutations of data."""

        # Create some random data
        np.random.seed(seed)
        n_costs = np.random.randint(1, 9)
        n_objectives = np.random.randint(1, 4)
        costs = np.random.randint(low=-1, high=1, size=(n_costs, n_objectives))

        # Get masks
        mask_distinct = algorithm(costs, distinct=True)
        ranks_non_distinct = algorithm(costs, distinct=False)

        # Permute the data
        permutation = np.random.permutation(np.arange(n_costs))
        assert np.min(mask_distinct) > 0
        assert np.min(ranks_non_distinct) > 0

        # When `distinct` is set to `False`, permutation invariance should hold
        assert np.all(ranks_non_distinct[permutation] == algorithm(costs[permutation], distinct=False))

        # Equally many should me marked in the mask, regardless of `distinct` or not
        assert np.sum(mask_distinct[permutation]) == np.sum(algorithm(costs[permutation], distinct=True))
        assert np.sum(ranks_non_distinct[permutation]) == np.sum(algorithm(costs[permutation], distinct=False))

    @pytest.mark.parametrize("seed, algorithm", itertools.product(seeds, paretorank_algorithms))
    def test_paretorank_on_simplex(self, seed, algorithm):
        """On a simplex every value has rank 1."""
        # Generate random data on a D-1 dimensional simplex
        np.random.seed(seed)
        n_costs = 99
        n_objectives = np.random.randint(2, 10)
        costs = generate_problem_simplex(n_costs, n_objectives)

        # Ranks are 1, 1, 1, 1, ...
        ranks = algorithm(costs, distinct=False)
        assert np.all(ranks == 1)

        # Ranks are 1, 1, 1, 1, ...
        ranks = algorithm(costs, distinct=True)
        assert np.all(ranks == 1)


class TestCrowdingDistance:
    def test_on_known_instance(self):
        """Test on a small instance computed by hand."""
        costs = np.array([[1, 2], [3, 9], [5, 3], [9, 1]], dtype=float)
        dist = crowding_distance(costs)
        assert np.all(dist == np.array([np.inf, np.inf, 0.8125, np.inf]))

    @pytest.mark.parametrize("seed", seeds)
    def test_invariance_under_permutations(self, seed):
        """Test that the algorithm in invariant under random permutations of data."""

        # Create some random data
        np.random.seed(seed)
        n_costs = np.random.randint(1, 9)
        n_objectives = np.random.randint(1, 4)

        # Permutation invariance does not hold if rows are duplicated
        # Therefore we use floats here
        costs = np.random.randn(n_costs, n_objectives)

        # Get masks
        distances = crowding_distance(costs)

        # Verify that one distance is returned for each row
        assert len(distances) == n_costs

        # Permute the data
        permutation = np.random.permutation(np.arange(n_costs))

        # Permutation invariance should hold
        assert np.all(distances[permutation] == crowding_distance(costs[permutation]))

    def test_no_side_effects(self):
        """Test that the input is not mutated."""

        # Generate random data
        np.random.seed(42)
        costs = np.random.randn(10, 3)

        # Copy the data before passing into function
        costs_before = costs.copy()
        crowding_distance(costs)

        # The input argument is not changed/mutated in any way
        assert costs_before.shape == costs.shape
        assert np.all(costs_before == costs)


if __name__ == "__main__":
    pytest.main(
        args=[".", "--doctest-modules", "--maxfail=5", "--cache-clear", "--color", "yes", "--durations", str(10)]
    )
